import copy
import logging
from enum import Enum

import numpy as np
from PIL import Image as PImage
from pyquaternion import Quaternion

from .utils import BaseGLTFStructure
from .materials import Material, Sampler
from .buffers import Vec3Array, Vec4Array, ScalarArray

logger = logging.getLogger(__name__)


class Primitive(BaseGLTFStructure):
    _positions = None
    _normals = None
    _indices = None
    _tangents = None
    mode = None
    material = None
    texcoords = None
    colors = None
    joints = None
    weights = None

    class Mode(Enum):
        POINTS = 0
        LINES = 1
        LINE_LOOP = 2
        LINE_STRIP = 3
        TRIANGLES = 4
        TRIANGLE_STRIP = 5
        TRIANGLE_FAN = 6

    def __init__(self, prim=None, gltf=None, name=None, joints=None, weights=None,
                 positions=None, normals=None, indices=None, texcoords=None, tangents=None,
                 material=None, image=None, texture=None, mode=None, colors=None):
        self.name = name
        if prim:
            if 'name' in prim:
                self.name = name
            if 'material' in prim:
                self.material = gltf.materials[prim.get('material')]
            if 'indices' in prim:
                self.indices = gltf.accessors[prim.get('indices')]
            if 'mode' in prim:
                self.mode = self.Mode(prim.get('mode'))
            attributes = prim.get('attributes')
            if attributes:
                if 'POSITION' in attributes:
                    self.positions = gltf.accessors[attributes.get('POSITION')]
                if 'NORMAL' in attributes:
                    self.normals = gltf.accessors[attributes.get('NORMAL')]
                if 'TANGENT' in attributes:
                    self.tangents = gltf.accessors[attributes.get('TANGENT')]
                self.texcoords = []
                if 'TEXCOORD_0' in attributes:
                    i = 0
                    while True:
                        texcoords = attributes.get('TEXCOORD_' + str(i))
                        if texcoords is None:
                            break
                        self.texcoords.append(gltf.accessors[texcoords])
                        i += 1
                self.colors = []
                if 'COLOR_0' in attributes:
                    i = 0
                    while True:
                        colors = attributes.get('COLOR_' + str(i))
                        if colors is None:
                            break
                        self.colors.append(gltf.accessors[colors])
                        i += 1
                self.joints = []
                if 'JOINTS_0' in attributes:
                    i = 0
                    while True:
                        joints = attributes.get('JOINTS_' + str(i))
                        if joints is None:
                            break
                        self.joints.append(gltf.accessors[joints])
                        i += 1
                self.weights = []
                if 'WEIGHTS_0' in attributes:
                    i = 0
                    while True:
                        weights = attributes.get('WEIGHTS_' + str(i))
                        if weights is None:
                            break
                        self.weights.append(gltf.accessors[weights])
                        i += 1
        else:
            self.positions = positions
            self.normals = normals
            self.indices = indices
            self.tangents = tangents
            self.material = material
            self.texcoords = texcoords if texcoords is not None else []
            self.colors = colors if colors is not None else []
            self.joints = joints if joints is not None else []
            self.weights = weights if weights is not None else []
            self.mode = mode and self.Mode[mode]
            if image:
                self.material = Material(image=image)
            if texture:
                self.material = Material(color_texture=texture)

    def __repr__(self):
        return self.name or super(Primitive, self).__repr__()

    def __contains__(self, item):
        return (self._positions == item or
                self._normals == item or
                self._indices == item or
                self._tangents == item or
                item in self.texcoords or
                item in self.colors or
                item in self.joints or
                item in self.weights)

    def sort_joints(self):
        if len(self.joints) > 1:
            joint_set_count = len(self.joints)
            joint_data = np.hstack(tuple(j.data for j in self.joints))
            weight_data = np.hstack(tuple(w.data for w in self.weights))
            sorted_weights = []
            sorted_joints = []
            for w, j in zip(weight_data, joint_data):
                pairs = sorted(zip(w, j), reverse=True)
                new_weights, new_joints = list(zip(*pairs))
                sorted_weights.append(new_weights)
                sorted_joints.append(new_joints)
            sorted_weights = np.array(sorted_weights, 'float32')
            sorted_joints = np.array(sorted_joints, 'uint16')
            self.weights = []
            self.joints = []
            for i in range(joint_set_count):
                self.weights.append(
                    Vec4Array(sorted_weights[:, i * 4:(i + 1) * 4].astype('float32'))
                )
                self.joints.append(
                    Vec4Array(sorted_joints[:, i * 4:(i + 1) * 4].astype('uint16'))
                )

    def split_transparency(self, alpha_mode=Material.AlphaMode.BLEND):
        if self.mode and self.mode != self.Mode.TRIANGLES:
            raise NotImplementedError('Prim splitting only implemented for triangles')

        if not self.material or not self.texcoords or (
                self.material.alpha_mode == Material.AlphaMode.OPAQUE):
            raise ValueError('Can\'t split prim: no material or texcoords, or material '
                             'is opaque, or material has no color_texture or diffuse_texture.')

        tex = self.material.color_texture or self.material.diffuse_texture
        if tex is None:
            return
        sampler = tex.sampler or Sampler()
        img = PImage.open(tex.source.get_fp())
        if len(img.getbands()) != 4:
            return
        alpha = img.getdata(3)
        min_alpha, max_alpha = alpha.getextrema()

        # return if all or none of the image is transparent
        if min_alpha == 255 or max_alpha < 255:
            return

        if self.material.color_uv is not None:
            texcoords = self.texcoords[self.material.color_uv]
        else:
            texcoords = self.texcoords[0]

        if self.indices:
            indices_iter = (self.indices.data[pos:pos + 3]
                            for pos in
                            range(0, self.indices.count, 3))
        else:
            indices_iter = ((i, i + 1, i + 2)
                            for i in
                            range(0, texcoords.count, 3))
        t_indices = []
        o_indices = []
        for indices in indices_iter:
            for point in [texcoords.data[i] for i in indices]:
                point = sampler.wrap_point(point)
                x = round((alpha.size[0] - 1) * point[0])
                y = round((alpha.size[1] - 1) * point[1])
                alpha_val = alpha.getpixel((x, y))
                if alpha_val < 255:
                    break
            else:
                o_indices.extend(indices)
                continue
            t_indices.extend(indices)

        # return if there are no opaque vertices
        if not len(o_indices):
            return

        t_material = copy.copy(self.material)
        t_material.alpha_mode = alpha_mode
        # Copy own material in case something else is using it
        self.material = copy.copy(self.material)
        self.material.alpha_mode = Material.AlphaMode.OPAQUE
        self.indices = o_indices

        return Primitive(positions=self.positions, normals=self.normals,
                         texcoords=self.texcoords, tangents=self.tangents,
                         joints=self.joints, weights=self.weights, colors=self.colors,
                         indices=t_indices, material=t_material)

    def render(self, gltf):
        primitive = {'attributes': {}}
        if self.mode:
            primitive['mode'] = self.mode.value
        if self.material:
            primitive['material'] = gltf.index('materials', self.material)
        if self.indices and self.indices.count:
            primitive['indices'] = gltf.index('accessors', self.indices)
        if self.positions and self.positions.count:
            primitive['attributes']['POSITION'] = gltf.index('accessors', self.positions)
        if self.normals and self.normals.count:
            primitive['attributes']['NORMAL'] = gltf.index('accessors', self.normals)
        if self.tangents and self.tangents.count:
            primitive['attributes']['TANGENT'] = gltf.index('accessors', self.tangents)
        for i, texcoords in enumerate(self.texcoords):
            primitive['attributes']['TEXCOORD_' + str(i)] = gltf.index('accessors', texcoords)
        for i, colors in enumerate(self.colors):
            primitive['attributes']['COLOR_' + str(i)] = gltf.index('accessors', colors)
        for i, joints in enumerate(self.joints):
            primitive['attributes']['JOINTS_' + str(i)] = gltf.index('accessors', joints)
        for i, weights in enumerate(self.weights):
            primitive['attributes']['WEIGHTS_' + str(i)] = gltf.index('accessors', weights)
        return primitive

    def repair(self):
        return
        if self.normals:
            magnitudes = np.sqrt((self.normals.data ** 2).sum(-1))[..., np.newaxis]

            if not np.all(np.isclose(magnitudes, 1)):
                self.normals.data /= magnitudes

            if np.any(magnitudes == 0):
                zero_indices, _ = np.where(magnitudes == 0)
                for index in zero_indices:
                    if self.indices:
                        matching = np.where(self.indices.data == index)[0]
                        if not matching.size:
                            # normal unused anyways
                            self.normals.data[index] = [0, 1, 0]
                            continue
                        index_in_indices = matching[0]
                        mod = index_in_indices % 3
                        a = self.positions.data[self.indices.data[index_in_indices - mod]]
                        b = self.positions.data[self.indices.data[1 + index_in_indices - mod]]
                        c = self.positions.data[self.indices.data[2 + index_in_indices - mod]]
                    else:
                        mod = index % 3
                        a = self.positions.data[index - mod]
                        b = self.positions.data[1 + index - mod]
                        c = self.positions.data[2 + index - mod]
                    dir = np.cross(b - a, c - a)
                    mag = np.linalg.norm(dir)
                    if mag == 0:
                        norm = [0, 1, 0]
                    else:
                        norm = dir / mag
                    self.normals.data[index] = norm

    @property
    def positions(self):
        return self._positions

    @positions.setter
    def positions(self, positions):
        if positions is None:
            self._positions = None
            return
        self._positions = positions if isinstance(positions, Vec3Array) else Vec3Array(positions)

    @property
    def normals(self):
        return self._normals

    @normals.setter
    def normals(self, normals):
        if normals is None:
            self._normals = None
            return
        self._normals = normals if isinstance(normals, Vec3Array) else Vec3Array(normals)

    @property
    def tangents(self):
        return self._tangents

    @tangents.setter
    def tangents(self, tangents):
        if tangents is None:
            self._tangents = None
            return
        self._tangents = tangents if isinstance(tangents, Vec4Array) else Vec4Array(tangents)

    @property
    def indices(self):
        return self._indices

    @indices.setter
    def indices(self, indices):
        if indices is None:
            self._indices = None
            return
        self._indices = indices if isinstance(indices, ScalarArray) else ScalarArray(indices)


class Mesh(BaseGLTFStructure):
    def __init__(self, mesh=None, gltf=None, name=None):
        self.primitives = []
        if mesh:
            self.name = mesh.get('name')
            for prim in mesh['primitives']:
                self.primitives.append(Primitive(prim, gltf))
        else:
            self.name = name

    def __repr__(self):
        return self.name or super(Mesh, self).__repr__()

    def __eq__(self, other):
        # Note that we don't consider `name` when merging nodes
        if not isinstance(other, Mesh):
            return False
        return self.primitives == other.primitives

    def add_primitive(self, **kwargs):
        self.primitives.append(Primitive(**kwargs))
        return len(self.primitives) - 1

    def render(self, gltf):
        mesh = {
            'primitives': [primitive.render(gltf) for primitive in self.primitives]
        }
        if self.name:
            mesh['name'] = self.name

        return mesh

    def repair(self):
        for p in self.primitives:
            p.repair()


class Camera(BaseGLTFStructure):
    perspective = None
    orthographic = None

    def __init__(self, camera=None, name=None, type=None, perspective=None, orthographic=None):
        if camera:
            self.name = camera.get('name')
            self.type = camera.get('type')
            self.perspective = camera.get('perspective')
            self.orthographic = camera.get('orthographic')
        else:
            self.name = name
            if type is None:
                if perspective is not None:
                    type = 'perspective'
                elif orthographic is not None:
                    type = 'orthographic'
                else:
                    raise ValueError('Camera must have type')
            self.type = type
            self.perspective = perspective
            self.orthographic = orthographic

    def render(self):
        camera = {}
        if self.name is not None:
            camera['name'] = self.name
        if self.type:
            camera['type'] = self.type
        if self.type == 'perspective':
            camera['perspective'] = self.perspective
        elif self.type == 'orthographic':
            camera['orthographic'] = self.orthographic
        return camera


class Node(BaseGLTFStructure):
    mesh = None
    translation = None
    rotation = None
    scale = None
    matrix = None
    skin = None
    camera = None

    def __init__(self, node=None, gltf=None, name=None, mesh=None, children=None,
                 skin=None, camera=None):
        if node:
            self.children = []
            self.name = node.get('name')

            if 'mesh' in node:
                self.mesh = gltf.meshes[node['mesh']]

            if 'camera' in node:
                self.camera = gltf.cameras[node['camera']]

            translation = node.get('translation')
            if translation:
                self.translation = np.array(translation, dtype='float32')

            rotation = node.get('rotation')
            if rotation:
                self.rotation = Quaternion(rotation[3], *rotation[:3])

            scale = node.get('scale')
            if scale:
                self.scale = np.array(scale, dtype='float32')

            matrix = node.get('matrix')
            if matrix:
                self.matrix = np.array(matrix, dtype='float32').reshape((4, 4))

            if matrix and (scale or rotation or translation):
                logger.warning('Node defined both matrix and at least one other transform.')
        else:
            self.name = name
            self.mesh = mesh
            self.children = children or []
            self.skin = skin
            self.camera = camera

    def __repr__(self):
        return self.name or super(Node, self).__repr__()

    def __eq__(self, other):
        if not (isinstance(other, type(self))
                and (self.mesh, self.name, self.children, self.skin, self.camera) ==
                (other.mesh, other.name, other.children, other.skin, other.camera)):
            return False

        # quaternions seem to blow up if they're compared to None
        if (self.rotation is None) != (other.rotation is None) or self.rotation != other.rotation:
            return False

        return (np.array_equal(self.translation, other.translation) and
                np.array_equal(self.scale, other.scale) and
                np.array_equal(self.matrix, other.matrix))

    def find_children(self, child_indices, gltf):
        for child_idx in child_indices:
            self.children.append(gltf.nodes[child_idx])

    @property
    def descendancy_has_mesh(self):
        if self.mesh:
            return True
        return any(c.descendancy_has_mesh for c in self.children)

    def get_transform(self, parent_transformation=None, parent_rotation=None):
        transformation = np.identity(4)
        rotation_matrix = parent_rotation
        if self.scale is not None:
            transformation = transformation.dot(
                np.diag(np.append(self.scale, [1]))
            )
        if self.rotation is not None:
            rotation_matrix = (
                self.rotation.inverse.rotation_matrix.dot(rotation_matrix)
                if rotation_matrix is not None
                else self.rotation.inverse.rotation_matrix
            )
            transformation = transformation.dot(
                self.rotation.inverse.transformation_matrix
            )
        if self.translation is not None:
            translation = np.identity(4)
            translation[3, :] += np.append(self.translation, [0])
            transformation = transformation.dot(translation)

        # Only use self.matrix if there is no other transform
        if self.matrix is not None and np.allclose(transformation, np.identity(4)):
            transformation = self.matrix

        if parent_transformation is not None:
            transformation = transformation.dot(parent_transformation)

            scale = [
                np.linalg.norm(transformation[:3, 0]),
                np.linalg.norm(transformation[:3, 1]),
                np.linalg.norm(transformation[:3, 2]),
            ]

            rotation_matrix = transformation[:3, :3] / scale

        return transformation, rotation_matrix

    def apply_transforms(self, parent_transformation=None, parent_rotation=None):
        if not self.descendancy_has_mesh:
            return

        transformation, rotation_matrix = self.get_transform(parent_transformation,
                                                             parent_rotation)

        if self.mesh and not np.allclose(transformation, np.identity(4)):
            for p in self.mesh.primitives:
                if not p.positions:
                    continue
                vertices = p.positions.data
                vertices = np.append(vertices, np.ones([len(vertices), 1]), 1)
                p.positions.data = vertices.dot(transformation)[:, :3].astype('float32')
                if p.normals and rotation_matrix is not None:
                    p.normals.data = p.normals.data.dot(rotation_matrix).astype('float32')

        for child in self.children:
            child.apply_transforms(transformation, rotation_matrix)

        self.translation = None
        self.rotation = None
        self.scale = None
        self.matrix = None

    def render(self, gltf):
        node = {}
        if self.mesh:
            node['mesh'] = gltf.index('meshes', self.mesh)
        if self.camera:
            node['camera'] = gltf.index('cameras', self.camera)
        if self.name:
            node['name'] = self.name
        if self.skin:
            node['skin'] = gltf.index('skins', self.skin)
        if self.children:
            node['children'] = [gltf.index_node(n) for n in self.children]
        if self.translation is not None:
            node['translation'] = self.translation.tolist()
        if self.rotation is not None:
            node['rotation'] = self.rotation.elements.tolist()
            node['rotation'].append(node['rotation'].pop(0))
        if self.scale is not None:
            node['scale'] = self.scale.tolist()
        if self.matrix is not None:
            node['matrix'] = self.matrix.reshape((16,)).tolist()
        return node



import os
import re
import math
import logging
import imghdr
import shutil
from enum import Enum
from uuid import uuid4
from io import BytesIO

import numpy as np
import requests
from PIL import Image as PImage

from .utils import BaseGLTFStructure, load_uri

logger = logging.getLogger(__name__)


class Image(object):
    name = ''
    _data = None
    mime_type = None
    gltf = None

    def __init__(self, image, gltf=None):
        if gltf:
            self.gltf = gltf

        if type(image) == dict:
            self.name = image.get('name', '')
            try:
                if 'bufferView' in image:
                    bv = gltf.buffer_views[image['bufferView']]
                    self.data = bv.data
                else:
                    self.data = image['uri']
            except KeyError:
                logger.warning('Image has no bufferView or uri')
        elif isinstance(image, PImage.Image):
            img_bytes = BytesIO()
            if len(image.getbands()) == 4:
                image.save(img_bytes, format='png')
            else:
                image.save(img_bytes, format='jpeg', quality=95)
            self.data = img_bytes.getvalue()
        else:
            self.data = image

    def __repr__(self):
        return self.name or super(Image, self).__repr__()

    def __eq__(self, other):
        if not isinstance(other, type(self)) or self.is_uri != other.is_uri:
            return False
        return (self.name, self.data) == (other.name, other.data)

    def __hash__(self):
        return hash(self.path or id(self))

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        mime_type = None
        try:
            str_type = unicode
        except NameError:
            str_type = str
        if type(data) is str_type:
            if data.startswith('data:'):
                data = load_uri(data)
            elif data.startswith('http'):
                if self.gltf.ignore_http_uri:
                    if '.jpg' in data or '.jpeg' in data:
                        mime_type = 'image/jpeg'
                    elif '.png' in data:
                        mime_type = 'image/png'
                else:
                    r = requests.get(data)
                    if r.status_code >= 400:
                        logger.warning('Unable to fetch image uri')
                    else:
                        mime_type = imghdr.what(None, r.content)
            else:
                data = data.replace('\\', '/')
                if os.path.isabs(data):
                    logger.warning('Image uri is an absolute path: ' + data)
                path = (os.path.join(self.gltf.folder, data)
                        if self.gltf and self.gltf.folder
                        else data)
                if os.path.exists(path):
                    mime_type = imghdr.what(path)
                else:
                    logger.warning('Image uri does not exist: ' + path)

        # Check type again in case a data uri was loaded above
        if type(data) is not str_type:
            mime_type = imghdr.what(None, data)
            if not mime_type:
                im = PImage.open(BytesIO(data))
                img_bytes = BytesIO()
                if len(im.getbands()) == 4:
                    im.save(img_bytes, format='PNG')
                else:
                    im.save(img_bytes, format='JPEG')

                data = img_bytes.getvalue()
                mime_type = imghdr.what(None, data)

        if mime_type:
            self.mime_type = 'image/' + mime_type
        else:
            logger.warning('Image format not recognized')

        self._data = data

    @property
    def is_uri(self):
        try:
            str_type = unicode
        except NameError:
            str_type = str
        return type(self._data) is str_type

    @property
    def path(self):
        if not self.is_uri:
            return
        if self.gltf and self.gltf.folder:
            return os.path.join(self.gltf.folder, self.data)
        return self.data

    def get_fp(self):
        if self.is_uri:
            with open(self.path, 'rb') as f:
                return BytesIO(f.read())

        return BytesIO(self.data)

    def set_uri_to(self, folder):
        data = self.data
        if self.is_uri:
            data = self.get_fp().getvalue()

        for file in os.listdir(folder):
            with open(os.path.join(folder, file), 'rb') as f:
                file_bytes = f.read()
                if file_bytes == data:
                    self.data = os.path.join(folder, file)
                    break
        else:
            self.data = self.write_to(folder)

    def write_to(self, folder, name=None):
        if self.is_uri:
            filename = name or os.path.basename(self.data)
            shutil.copy(self.path, os.path.join(folder, filename))
            return filename

        filename = name or self.name or str(uuid4())[:8]
        if '.' in filename:
            ext = '.' + filename.split('.')[-1]
            filename = '.'.join(filename.split('.')[:-1])
        else:
            ext = '.'
            if self.mime_type == 'image/jpeg':
                ext += 'jpg'
            else:
                ext += self.mime_type.split('/')[-1]

        clean_filename = str(filename + ext).strip().replace(' ', '_')
        clean_filename = re.sub(r'(?u)[^-\w.]', '', clean_filename)

        path = os.path.join(folder, clean_filename)
        if os.path.exists(path) and name is None:
            filename = str(uuid4())[:8]
            path = os.path.join(folder, filename + ext)

        with open(path, 'wb') as f:
            f.write(self.data)

        return path

    def render(self, buffer, gltf, embed=False, write_to=None, splitting=False):
        image = {}
        if self.name:
            image['name'] = self.name

        data = self.data

        if not data:
            return image
        if self.mime_type is not None:
            image['mimeType'] = self.mime_type

        if write_to:
            folder = os.path.dirname(write_to)
            name = os.path.basename(write_to)
            image['uri'] = self.write_to(folder, name + str(gltf.images.index(self)))
            if splitting:
                image['uri'] = os.path.basename(image['uri'])
            return image

        if self.is_uri:
            if embed:
                try:
                    data = load_uri(self.data, folder=gltf.folder)
                except FileNotFoundError:
                    logger.error('Unable to load image uri: ' + self.data)
                    img_bytes = BytesIO()
                    PImage.new('RGB', (1, 1)).save(img_bytes, 'JPEG')
                    data = img_bytes.getvalue()
            else:
                image['uri'] = self.data
                return image

        view = buffer.get_view(image=self)
        if view.data:
            if view.byte_length != len(data):
                raise ValueError('Buffer view already exists for this image, '
                                 'but the byte_length does not match')
        else:
            view.write(data)
        image['bufferView'] = gltf.index('buffer_views', view)
        return image


class Sampler(BaseGLTFStructure):
    mag_filter = None
    min_filter = None
    wrap_s = None
    wrap_t = None

    class MagFilter(Enum):
        NEAREST = 9728
        LINEAR = 9729

    class MinFilter(Enum):
        NEAREST = 9728
        LINEAR = 9729
        NEAREST_MIPMAP_NEAREST = 9984
        LINEAR_MIPMAP_NEAREST = 9985
        NEAREST_MIPMAP_LINEAR = 9986
        LINEAR_MIPMAP_LINEAR = 9987

    class WrapMode(Enum):
        CLAMP_TO_EDGE = 33071
        MIRRORED_REPEAT = 33648
        REPEAT = 10497

    def __init__(self, sampler=None, name=None,
                 mag_filter=None, min_filter=None, wrap_s=None, wrap_t=None):
        if sampler:
            mag_filter = sampler.get('magFilter', mag_filter)
            if mag_filter:
                self.mag_filter = self.MagFilter(mag_filter)
            min_filter = sampler.get('minFilter', min_filter)
            if min_filter:
                self.min_filter = self.MinFilter(min_filter)
            wrap_s = sampler.get('wrapS', wrap_s)
            if wrap_s:
                self.wrap_s = self.WrapMode(wrap_s)
            wrap_t = sampler.get('wrapT', wrap_t)
            if wrap_t:
                self.wrap_t = self.WrapMode(wrap_t)
            self.name = sampler.get('name')
        else:
            self.name = name
            self.mag_filter = mag_filter
            self.min_filter = min_filter
            self.wrap_s = wrap_s
            self.wrap_t = wrap_t

    def __repr__(self):
        return self.name or super(Sampler, self).__repr__()

    def render(self):
        sampler = {}
        if self.name:
            sampler['name'] = self.name
        if self.mag_filter:
            sampler['magFilter'] = self.mag_filter.value
        if self.min_filter:
            sampler['minFilter'] = self.min_filter.value
        if self.wrap_s:
            sampler['wrapS'] = self.wrap_s.value
        if self.wrap_t:
            sampler['wrapT'] = self.wrap_t.value
        return sampler

    @staticmethod
    def wrap(val, wrap_mode):
        if 0 <= val <= 1:
            return val
        if wrap_mode == Sampler.WrapMode.CLAMP_TO_EDGE:
            return 1.0 if val > 1 else 0.0
        if wrap_mode == Sampler.WrapMode.REPEAT:
            portion, _ = math.modf(val)
            return portion if val > 0 else 1.0 + portion
        if wrap_mode == Sampler.WrapMode.MIRRORED_REPEAT:
            portion, whole = math.modf(abs(val))
            if (whole + 1) % 2:
                return portion
            return 1.0 - portion
        raise TypeError('Invalid WrapMode')

    def wrap_point(self, point):
        return (
            self.wrap(point[0], self.wrap_s or self.WrapMode.REPEAT),
            self.wrap(point[1], self.wrap_t or self.WrapMode.REPEAT)
        )


class Texture(BaseGLTFStructure):
    sampler = None
    source = None

    def __init__(self, texture=None, gltf=None, name=None,
                 image=None, sampler=None, mirrored_repeat=False, repeat=False):
        if texture:
            sam = texture.get('sampler')
            if sam is not None:
                self.sampler = sam if isinstance(sam, Sampler) else gltf.samplers[sam]
            img = texture.get('source')
            if img is not None:
                self.source = img if isinstance(img, Image) else gltf.images[img]
            self.name = texture.get('name')
        else:
            self.name = name
            self.sampler = sampler
            if mirrored_repeat:
                self.sampler = Sampler(wrap_s=Sampler.WrapMode.MIRRORED_REPEAT,
                                       wrap_t=Sampler.WrapMode.MIRRORED_REPEAT)
            if repeat:
                self.sampler = Sampler(wrap_s=Sampler.WrapMode.REPEAT,
                                       wrap_t=Sampler.WrapMode.REPEAT)
            if image:
                self.source = image if isinstance(image, Image) else Image(image)

    def __repr__(self):
        return self.name or super(Texture, self).__repr__()

    def render(self, gltf):
        texture = {'source': gltf.index('images', self.source)}
        if self.name:
            texture['name'] = self.name
        if self.sampler:
            texture['sampler'] = gltf.index('samplers', self.sampler)
        return texture


class Material(BaseGLTFStructure):
    alpha_mode = None
    double_sided = None
    base_color_factor = None
    color_texture = None
    color_uv = None
    color_extensions = None
    rough_metal_texture = None
    rough_uv = None
    rough_metal_extensions = None
    normal_texture = None
    normal_uv = None
    normal_scale = None
    normal_extensions = None
    occlusion_texture = None
    occlusion_uv = None
    occlusion_strength = None
    occlusion_extensions = None
    emissive_texture = None
    emissive_factor = None
    emissive_uv = None
    emissive_extensions = None
    metallic_factor = None
    roughness_factor = None

    diffuse_factor = None
    specular_factor = None
    glossiness_factor = None
    diffuse_texture = None
    spec_gloss_texture = None

    class AlphaMode(Enum):
        OPAQUE = 'OPAQUE'
        MASK = 'MASK'
        BLEND = 'BLEND'

    def __init__(self, material=None, gltf=None, name=None, image=None,
                 base_color_factor=None, double_sided=None,
                 color_texture=None, color_uv=None,
                 normal_texture=None, normal_uv=None,
                 occlusion_texture=None, occlusion_uv=None,
                 emissive_texture=None, emissive_uv=None,
                 rough_metal_texture=None, rough_uv=None,
                 alpha_mode=None, metallic_factor=0, roughness_factor=None, emissive_factor=None,
                 diffuse_factor=None, specular_factor=None, glossiness_factor=None,
                 diffuse_texture=None, spec_gloss_texture=None):
        if material:
            self.name = material.get('name')
            if material.get('alphaMode'):
                self.alpha_mode = self.AlphaMode(material.get('alphaMode'))
            if material.get('doubleSided'):
                self.double_sided = material.get('doubleSided')
            self.emissive_factor = material.get('emissiveFactor')
            pbr = material.get('pbrMetallicRoughness')
            if pbr:
                self.base_color_factor = tuple(pbr.get('baseColorFactor', []))
                self.metallic_factor = pbr.get('metallicFactor')
                self.roughness_factor = pbr.get('roughnessFactor')

                base_color_texture = pbr.get('baseColorTexture')
                if base_color_texture:
                    self.color_texture = gltf.textures[base_color_texture['index']]
                    if 'texCoord' in base_color_texture:
                        self.color_uv = base_color_texture['texCoord']
                    if 'extensions' in base_color_texture:
                        self.color_extensions = base_color_texture['extensions']

                metallic_roughness_tex = pbr.get('metallicRoughnessTexture')
                if metallic_roughness_tex:
                    self.rough_metal_texture = gltf.textures[metallic_roughness_tex['index']]
                    if 'texCoord' in metallic_roughness_tex:
                        self.rough_uv = metallic_roughness_tex['texCoord']
                    if 'extensions' in metallic_roughness_tex:
                        self.rough_metal_extensions = metallic_roughness_tex['extensions']

            normal = material.get('normalTexture')
            if normal:
                if 'index' in normal:
                    self.normal_texture = gltf.textures[normal['index']]
                if 'texCoord' in normal:
                    self.normal_uv = normal['texCoord']
                if 'scale' in normal:
                    self.normal_scale = normal['scale']
                if 'extensions' in normal:
                    self.normal_extensions = normal['extensions']

            occlusion = material.get('occlusionTexture')
            if occlusion:
                if 'index' in occlusion:
                    self.occlusion_texture = gltf.textures[occlusion['index']]
                if 'texCoord' in occlusion:
                    self.occlusion_uv = occlusion['texCoord']
                if 'strength' in occlusion:
                    self.occlusion_strength = occlusion['strength']
                if 'extensions' in occlusion:
                    self.occlusion_extensions = occlusion['extensions']

            emissive = material.get('emissiveTexture')
            if emissive:
                if 'index' in emissive:
                    self.emissive_texture = gltf.textures[emissive['index']]
                if 'texCoord' in emissive:
                    self.emissive_uv = emissive['texCoord']
                if 'extensions' in emissive:
                    self.emissive_extensions = emissive['extensions']

            extensions = material.get('extensions')
            if extensions:
                spec_gloss = extensions.get('KHR_materials_pbrSpecularGlossiness')
                if spec_gloss:
                    self.diffuse_factor = spec_gloss.get('diffuseFactor')
                    self.specular_factor = spec_gloss.get('specularFactor')
                    self.glossiness_factor = spec_gloss.get('glossinessFactor')
                    diffuse_texture = spec_gloss.get('diffuseTexture')
                    if diffuse_texture and diffuse_texture.get('index') is not None:
                        self.diffuse_texture = gltf.textures[diffuse_texture['index']]
                    spec_gloss_texture = spec_gloss.get('specularGlossinessTexture')
                    if spec_gloss_texture and spec_gloss_texture.get('index') is not None:
                        self.spec_gloss_texture = gltf.textures[spec_gloss_texture['index']]
        else:
            self.name = name
            self.double_sided = double_sided
            self.base_color_factor = base_color_factor
            self.color_texture = color_texture
            self.color_uv = color_uv
            self.normal_texture = normal_texture
            self.normal_uv = normal_uv
            self.occlusion_texture = occlusion_texture
            self.occlusion_uv = occlusion_uv
            self.rough_metal_texture = rough_metal_texture
            self.rough_uv = rough_uv
            self.emissive_texture = emissive_texture
            self.emissive_uv = emissive_uv
            self.metallic_factor = metallic_factor
            self.roughness_factor = roughness_factor
            self.emissive_factor = emissive_factor
            self.diffuse_factor = diffuse_factor
            self.specular_factor = specular_factor
            self.glossiness_factor = glossiness_factor
            self.diffuse_texture = diffuse_texture
            self.spec_gloss_texture = spec_gloss_texture
            if alpha_mode:
                self.alpha_mode = self.AlphaMode(alpha_mode)
            if image:
                self.color_texture = Texture(image=image)

    def __repr__(self):
        return self.name or super(Material, self).__repr__()

    def __hash__(self):
        return id(self)

    def __contains__(self, item):
        return (self.color_texture == item or
                self.rough_metal_texture == item or
                self.normal_texture == item or
                self.emissive_texture == item or
                self.diffuse_texture == item or
                self.spec_gloss_texture == item)

    def render(self, gltf):
        material = {'pbrMetallicRoughness': {}}

        if self.color_texture:
            color_texture = {
                'index': gltf.index('textures', self.color_texture)
            }
            if self.color_uv is not None:
                color_texture['texCoord'] = self.color_uv
            if self.color_extensions:
                color_texture['extensions'] = self.color_extensions
            material['pbrMetallicRoughness']['baseColorTexture'] = color_texture

        if self.rough_metal_texture:
            rough_metal_texture = {
                'index': gltf.index('textures', self.rough_metal_texture)
            }
            if self.rough_uv is not None:
                rough_metal_texture['texCoord'] = self.rough_uv
            if self.rough_metal_extensions:
                rough_metal_texture['extensions'] = self.rough_metal_extensions
            material['pbrMetallicRoughness']['metallicRoughnessTexture'] = rough_metal_texture

        if self.normal_texture:
            normal_texture = {
                'index': gltf.index('textures', self.normal_texture)
            }
            if self.normal_uv is not None:
                normal_texture['texCoord'] = self.normal_uv
            if self.normal_scale is not None:
                normal_texture['scale'] = self.normal_scale
            if self.normal_extensions:
                normal_texture['extensions'] = self.normal_extensions
            material['normalTexture'] = normal_texture

        if self.occlusion_texture:
            occlusion_texture = {
                'index': gltf.index('textures', self.occlusion_texture)
            }
            if self.occlusion_uv is not None:
                occlusion_texture['texCoord'] = self.occlusion_uv
            if self.occlusion_strength is not None:
                occlusion_texture['strength'] = self.occlusion_strength
            if self.occlusion_extensions:
                occlusion_texture['extensions'] = self.occlusion_extensions
            material['occlusionTexture'] = occlusion_texture

        if self.emissive_texture:
            emissive_texture = {
                'index': gltf.index('textures', self.emissive_texture)
            }
            if self.emissive_uv is not None:
                emissive_texture['texCoord'] = self.emissive_uv
            if self.emissive_extensions:
                emissive_texture['extensions'] = self.emissive_extensions
            material['emissiveTexture'] = emissive_texture

        if self.name:
            material['name'] = self.name
        if self.alpha_mode:
            material['alphaMode'] = self.alpha_mode.value
        if self.double_sided:
            material['doubleSided'] = self.double_sided
        if self.emissive_factor:
            material['emissiveFactor'] = self.emissive_factor
        if self.metallic_factor is not None:
            material['pbrMetallicRoughness']['metallicFactor'] = self.metallic_factor
        if self.roughness_factor is not None:
            material['pbrMetallicRoughness']['roughnessFactor'] = self.roughness_factor
        if self.base_color_factor:
            material['pbrMetallicRoughness']['baseColorFactor'] = self.base_color_factor

        spec_gloss = {}
        if self.diffuse_texture:
            spec_gloss['diffuseTexture'] = {'index': gltf.index('textures', self.diffuse_texture)}
        if self.spec_gloss_texture:
            spec_gloss['specularGlossinessTexture'] = {
                'index': gltf.index('textures', self.spec_gloss_texture)
            }
        if self.glossiness_factor is not None:
            spec_gloss['glossinessFactor'] = self.glossiness_factor
        if self.diffuse_factor:
            spec_gloss['diffuseFactor'] = self.diffuse_factor
        if self.specular_factor:
            spec_gloss['specularFactor'] = self.specular_factor
        if spec_gloss:
            material['extensions'] = {'KHR_materials_pbrSpecularGlossiness': spec_gloss}
            if 'KHR_materials_pbrSpecularGlossiness' not in gltf.extensions_used:
                gltf.extensions_used.append('KHR_materials_pbrSpecularGlossiness')

        if not material['pbrMetallicRoughness']:
            material.pop('pbrMetallicRoughness')
        return material

    def repair(self):
        if self.emissive_factor and isinstance(self.emissive_factor, list):
            new_emissive = []
            for f in self.emissive_factor:
                if type(f) is str:
                    f = float(f)
                if f < 0:
                    f = 0
                if f > 1:
                    f = 1
                new_emissive.append(f)
            self.emissive_factor = new_emissive
        if self.base_color_factor and isinstance(self.base_color_factor, list):
            new_color = []
            for f in self.base_color_factor:
                if type(f) is str:
                    f = float(f)
                if f < 0:
                    f = 0
                if f > 1:
                    f = 1
                new_color.append(f)
            self.base_color_factor = new_color

        # Fix VALUE_NOT_IN_RANGE errors
        if self.roughness_factor:
            self.roughness_factor = np.clip(self.roughness_factor, 0, 1.0)
        if self.metallic_factor:
            self.metallic_factor = np.clip(self.metallic_factor, 0, 1.0)
        if self.normal_scale:
            self.normal_scale = np.clip(self.normal_scale, -1.0, 1.0)



"""
Easily load and operate on GLTF files with this library.

```python
from gltf import GLTF

model = GLTF.load("input.glb")
model.repair()
model.save("output.glb")
```

Useful operations exist in the `ops` module.
"""

from .gltf import GLTF, Scene
from .animations import Animation, Skin
from .buffers import Buffer, BufferView, Vec2Array, Vec3Array, Vec4Array, ScalarArray, Mat4Array
from .materials import Material, Texture, Sampler, Image
from .nodes import Node, Mesh, Primitive



#!/usr/bin/env python

from setuptools import setup

VERSION = '0.5.15'

DESCRIPTION = """GLTF provides the ability to load, modify, and save GLTF/GLB files.

[GLTF](https://www.khronos.org/gltf/) is an open 3D model standard by the Khronos Group.
"""

DEPENDENCIES = ['numpy', 'pyquaternion', 'pillow', 'requests']

setup(
    name='gltf',
    version=VERSION,
    description="Load, modify, and save GLTF files.",
    long_description=DESCRIPTION,
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Multimedia :: Graphics :: 3D Modeling',
    ],
    author='Trey Nelson',
    author_email='trey@seekxr.com',
    license="Public Domain",
    packages=['gltf'],
    install_requires=DEPENDENCIES,
    zip_safe=False,
    scripts=['bin/gltf-op'],
)

